# Copyright (c) Chai Nunes 2016

class __MethodDict(dict):
    def __init__(self, **kwargs):
        dict.__init__(kwargs)
        for key, value in kwargs.items():
            setattr(self, key, value)
        
    def __call__(self, name, *args):
        self[name](*args)

class __StructType(object):
    def __init__(self, **kwargs):
        for key, value in kwargs.items():
            setattr(self, key, value)
        self.methods = __MethodDict()

    def __setattr__(self, key, value):
        if hasattr(self, key) or key == 'methods':
            super(__StructType, self).__setattr__(key, value)
        else:
            getattr(self, key) # Trigger AttributeError

    def __repr__(self):
        return repr(self.__dict__)

    def addattr(self, key, value=None):
        ret = Struct(self.__name__, (' '.join(self.__dict__.keys()) + " " + key))()
        for field in ' '.join(self.__dict__.keys()):
            setattr(ret, field, getattr(self, field))
        self.__class__ = ret.__class__
        setattr(self, key, value)

    def addCallback(self, name, callback):
        if isinstance(name, str):
            self.methods.update({name: callback})
            setattr(self.methods, name, callback)

    def setCallback(self, name, callback):
        self.methods[name] = callback
        setattr(self.methods, name, callback)

    def getCallback(self, name):
        return self.methods[name]

    def delCallback(self, name):
        del self.methods[name]
        delattr(self.methods, name)
            
    def __call__(self, name, *args):
        self.methods(name, *args)

def Struct(typename, field_names=""):
    return type(typename, (__StructType,), dict.fromkeys(field_names.split() if isinstance(field_names, str) else field_names))
